#include "precompile.h"

#include <float.h>

#include "playermgr.h"
#include "player.h"
#include "cs_proto.pb.h"
#include "room.h"
#include "metamgr.h"
#include "loot.h"
#include "app.h"
#include "collider.h"
#include "typeconvert.h"

Player::Player():Human()
{
    entity_type = ET_Player;
    entity_subtype = EST_Player;
    ++App::Instance()->perf.entity_num[ET_Player];
}

Player::~Player()
{
    --App::Instance()->perf.entity_num[ET_Player];
}

void Player::Initialize()
{
    Human::Initialize();
    RecalcBuff();
}

void Player::Update(int delta_time)
{
    if (moving) {
        UpdateMove();
    }
    if (room->frame_no % 2 == 0) {
        if (spectate) {
            UpdateSpectate();
        }
        if (emote) {
            UpdateEmote();
        }
    }
}

void Player::UpdateMove()
{
    if (dead) {
        moving = false;
        moved_frames = 0;
        return;
    }
    ++moved_frames;
    if (moved_frames > 4) {
        moving = false;
        moved_frames = 0;
        return;
    }
    _UpdateMove(std::max(1, (int)GetSpeed()));
}

void Player::UpdateSpectate()
{
    if (room->gas_data.gas_mode == GasInactive) {
        spectate = false;
        return;
    }
    BeKill(entity_uniid, name);
    spectate = false;
}

void Player::UpdateEmote()
{
    room->frame_event.AddEmote(this, emote_id);
    emote = false;
    emote_id = 0;
}

void Player::_CMMove(f8::MsgHdr& hdr, const cs::CMMove& msg)
{
    moving = false;
    if (msg.has_move_dir()) {
        if (std::abs(msg.move_dir().x()) > FLT_EPSILON ||
            std::abs(msg.move_dir().y()) > FLT_EPSILON
            ) {
            a8::Vec2 old_move_dir;
            TypeConvert::FromPb(move_dir, &msg.move_dir());
            move_dir.Normalize();
            moving = true;
        }
    }
    assert(!isnan(move_dir.x) && !isnan(move_dir.y));
    if (moving) {
        moved_frames = 0;
    }
    last_seq_id = msg.seq();
    if (msg.has_spectate()) {
        spectate = true;
    }
    if (msg.has_emote()) {
        emote = true;
        emote_id = msg.emote();
    }
}

void Player::_CMEmote(f8::MsgHdr& hdr, const cs::CMEmote& msg)
{

}

void Player::_CMVoice(f8::MsgHdr& hdr, const cs::CMVoice& msg)
{
    cs::SMVoiceNotify notifymsg;
    notifymsg.set_account_id(account_id);
    notifymsg.set_download_url(msg.download_url());
    auto send_func = [] (Player* hum, a8::XParams& param)
                     {
                         cs::SMVoiceNotify* msg = (cs::SMVoiceNotify*)param.sender.GetUserData();
                         hum->SendNotifyMsg(*msg);
                     };
    room->TouchPlayerList(a8::XParams()
                          .SetSender(&notifymsg),
                          send_func);
}

void Player::_CMGameOver(f8::MsgHdr& hdr, const cs::CMGameOver& msg)
{
    SendGameOver();
}

void Player::_CMWatchWar(f8::MsgHdr& hdr, const cs::CMWatchWar& msg)
{
    cs::SMWatchWar respmsg;
    Human* target = room->GetWatchWarTarget(this);
    if (target) {
        SendNotifyMsg(respmsg);
        hdr.hum->FollowTarget(target);
    } else {
        respmsg.set_error_code(1);
        SendNotifyMsg(respmsg);
    }
}

void Player::_CMLeave(f8::MsgHdr& hdr, const cs::CMLeave& msg)
{
    if (!leave_) {
        if (room->BattleStarted()) {
            //如果战斗已经开始算自杀
            BeKill(entity_uniid, name);
        }
        leave_ = true;
        leave_frameno_ = room->frame_no;
    }
    cs::SMLeave respmsg;
    SendNotifyMsg(respmsg);
}

void Player::_CMRevive(f8::MsgHdr& hdr, const cs::CMRevive& msg)
{
    if (real_dead) {
        SendGameOver();
        return;
    }
    if (dead && !real_dead && revive_timer) {
        dead = false;
        SyncAroundPlayers(__FILE__, __LINE__, __func__);
        room->xtimer.DeleteTimer(revive_timer);
        revive_timer = nullptr;
    }
}
