#include "precompile.h"

#include <random>

#include <a8/mutable_xobject.h>
#include <a8/timer.h>
#include <a8/udplog.h>
#include <a8/collision.h>

#include "playermgr.h"
#include "player.h"
#include "cs_proto.pb.h"
#include "room.h"
#include "android.h"
#include "metamgr.h"
#include "collider.h"
#include "loot.h"
#include "roommgr.h"
#include "app.h"
#include "gamelog.h"
#include "typeconvert.h"

const int MAP_CELL_WIDTH = 64 * 8;
const int MAP_GRID_WIDTH = 64;
const int ROOM_MAX_PLAYER_NUM = 40;

static long long RoomXGetTickCount(void* context)
{
    Room* room = (Room*)context;
    return room->frame_no;
}

Room::~Room()
{

}

void Room::Init()
{
    if (map_meta->p->map_width() < 1) {
        abort();
    }
    if (map_meta->p->map_height() < 1) {
        abort();
    }

    xtimer.Init(RoomXGetTickCount, this, 100, 100);
    xtimer_attacher.xtimer = &xtimer;
    grid_service.Init(map_meta->p->map_width(),
                      map_meta->p->map_height(),
                      MAP_CELL_WIDTH);
    map_service.Init(map_meta->p->map_width() / MAP_GRID_WIDTH,
                     map_meta->p->map_height() / MAP_GRID_WIDTH,
                     MAP_GRID_WIDTH);

    ShuaAndroid();
}

void Room::UnInit()
{
    App::Instance()->perf.alive_count -= alive_count_;
    if (game_over_timer) {
        a8::Timer::Instance()->DeleteTimer(game_over_timer);
        game_over_timer = nullptr;
    }
    xtimer_attacher.ClearTimerList();
    for (auto& pair : accountid_hash_) {
        PlayerMgr::Instance()->RemovePlayerBySocket(pair.second->socket_handle);
    }
    for (auto& pair : uniid_hash_) {
        delete pair.second;
    }
    uniid_hash_.clear();
    for (auto& pair : later_add_hash_) {
        delete pair.second;
    }
    later_add_hash_.clear();
    for (auto& pair : removed_robot_hash_) {
        delete pair.second;
    }
    removed_robot_hash_.clear();
    grid_service.UnInit();
    map_service.UnInit();
}

void Room::Update(int delta_time)
{
    xtimer.Update();
    if (game_over && frame_no - game_over_frameno > SERVER_FRAME_RATE * 20) {
        return;
    }

    elapsed_time_ += delta_time;
    while (elapsed_time_ >= 50) {
        if (frame_no % 2 == 0) {
            UpdateGas();
        }
        for (auto& pair : moveable_hash_) {
            if (pair.second->entity_type == ET_Player &&
                pair.second->updated_times <= 0) {
                if (frame_no % 2 == 0) {
                    pair.second->Update(50);
                    pair.second->updated_times++;
                }
            } else {
                pair.second->Update(50);
                pair.second->updated_times++;
            }
        }
        if (frame_no % 2 == 0) {
            for (auto& pair : human_hash_) {
                pair.second->SendUpdateMsg();
            }
            frame_event.Clear();
        }
        ++frame_no;
        elapsed_time_ -= 50;
    }
}

int Room::GetPlayerNum()
{
    return accountid_hash_.size();
}

Player* Room::GetPlayerByAccountId(const std::string& accountid)
{
    auto itr = accountid_hash_.find(accountid);
    return itr != accountid_hash_.end() ? itr->second : nullptr;
}

Player* Room::GetPlayerByUniId(int uniid)
{
    Entity* entity = GetEntityByUniId(uniid);
    return entity && entity->entity_type == ET_Player && entity->entity_subtype == EST_Player ? (Player*)entity : nullptr;
}

Entity* Room::GetEntityByUniId(int uniid)
{
    auto itr = uniid_hash_.find(uniid);
    return itr != uniid_hash_.end() ? itr->second : nullptr;
}

int Room::AliveCount()
{
    return alive_count_;
}

void Room::AddPlayer(Player* hum)
{
    assert(gas_data.gas_mode == GasInactive);
    {
        hum->pos.x = 100;
        hum->pos.y = 100;
    }
    hum->entity_uniid = AllocUniid();
    hum->room = this;
    hum->join_frameno = frame_no;
    hum->Initialize();
    uniid_hash_[hum->entity_uniid] = hum;
    moveable_hash_[hum->entity_uniid] = hum;
    accountid_hash_[hum->account_id] = hum;
    human_hash_[hum->entity_uniid] = hum;
    ++alive_count_;
    ++App::Instance()->perf.alive_count;
    grid_service.AddHuman(hum);
    hum->RefreshView();
    MatchTeam(hum);
    while (human_hash_.size() > ROOM_MAX_PLAYER_NUM) {
        RandRemoveAndroid();
    }
    NotifyUiUpdate();
}

int Room::AllocUniid()
{
    while (GetEntityByUniId(++current_uniid) ||
           later_add_hash_.find(current_uniid) != later_add_hash_.end() ||
           current_uniid == 0) {}
    return current_uniid;
}

void Room::ShuaAndroid()
{
    if (gas_data.gas_mode != GasInactive) {
        return;
    }
    int robot_min_num = MetaMgr::Instance()->GetSysParamAsInt("refresh_robot_min_num", 5);
    int robot_max_num = MetaMgr::Instance()->GetSysParamAsInt("refresh_robot_max_num", 10);
    int refresh_min_time = MetaMgr::Instance()->GetSysParamAsInt("refresh_robot_min_time", 5);
    int refresh_max_time = MetaMgr::Instance()->GetSysParamAsInt("refresh_robot_max_time", 10);
    int robot_num = a8::RandEx(robot_min_num, robot_max_num);
    int refresh_time = a8::RandEx(refresh_min_time, refresh_max_time);
    if (robot_num > 0 && refresh_time > 0) {
        CreateAndroid(robot_num);
        xtimer.AddDeadLineTimerAndAttach(SERVER_FRAME_RATE * refresh_time,
                                         a8::XParams()
                                         .SetSender(this),
                                         [] (const a8::XParams& param)
                                         {
                                             Room* room = (Room*)param.sender.GetUserData();
                                             room->ShuaAndroid();
                                         },
                                         &xtimer_attacher.timer_list_);
        NotifyUiUpdate();
    }
}

void Room::CreateAndroid(int robot_num)
{
    #if 0
    std::vector<MetaData::Robot>* robot_list = MetaMgr::Instance()->GetRobotList();
    if (!robot_list || robot_list->empty()) {
        return;
    }
    MetaData::Player* hum_meta = MetaMgr::Instance()->GetPlayer(40002);
    assert(hum_meta);
    if (!hum_meta) {
        abort();
    }
    for (int i = 0; i < robot_num; ++i) {
        if (human_hash_.size() >= ROOM_MAX_PLAYER_NUM) {
            return;
        }

        MetaData::Robot* robot_meta = nullptr;
        int try_count = 0;
        while (true) {
            MetaData::Robot& tmp_robot_meta = (*robot_list)[rand() % robot_list->size()];
            if (refreshed_robot_set_.find(tmp_robot_meta.p->id()) ==
                refreshed_robot_set_.end()) {
                robot_meta = &tmp_robot_meta;
                break;
            }
            ++try_count;
            if (try_count > 3000) {
                return;
            }
        }

        Android* hum = new Android();
        hum->name = robot_meta->p->name();
        hum->meta = hum_meta;
        hum->robot_meta = robot_meta;
        hum->entity_uniid = AllocUniid();
        hum->attack_dir = hum->pos;
        hum->attack_dir.Normalize();
        hum->attack_dir.Rotate(a8::RandAngle());
        hum->move_dir = hum->attack_dir;

        hum->room = this;
        hum->Initialize();
        uniid_hash_[hum->entity_uniid] = hum;
        moveable_hash_[hum->entity_uniid] = hum;
        human_hash_[hum->entity_uniid] = hum;
        ++alive_count_;
        ++App::Instance()->perf.alive_count;
        grid_service.AddHuman(hum);
        hum->FindLocation();
        hum->RefreshView();
        {
            hum->team_id = NewTeam();
            hum->team_members = &team_hash_[hum->team_id];
            hum->team_members->insert(hum);
        }
        refreshed_robot_set_.insert(robot_meta->p->id());
    }
    #endif
}

Human* Room::FindEnemy(Human* hum)
{
    std::vector<Human*> enemys;
    enemys.reserve(50);
    EntitySubType_e sub_type = EST_Player;
    if ((rand() % 10) < 2) {
        sub_type = EST_Android;
    }
    for (auto& cell : hum->grid_list) {
        for (Human* target : cell->human_list) {
            if (target->entity_subtype == sub_type &&
                !target->dead) {
                if (hum->pos.Distance(target->pos) < 300.0f) {
                    if (target->team_id == 0 ||
                        target->team_id != hum->team_id
                        ) {
                        enemys.push_back(target);
                    }
                }
            }
        }
    }
    std::sort(enemys.begin(), enemys.end(),
              [hum] (Human* a, Human *b) -> bool
              {
                  return (hum->pos - a->pos).Norm() < (hum->pos - b->pos).Norm();
              });
    return !enemys.empty() ? enemys[0] : nullptr;
}

void Room::FillSMJoinedNotify(Player* self_hum, cs::SMJoinedNotify& msg)
{
    msg.set_team_mode(msg.team_mode());
    msg.set_player_id(self_hum->entity_uniid);
    msg.set_started(false);
    msg.set_room_uuid(a8::XValue(room_uuid).GetString());
}

int Room::CreateLoot(int equip_id, a8::Vec2 pos, int count, int equip_lv)
{
    return 0;
}

void Room::RemoveObjectLater(Entity* entity)
{
    auto remove_func = [] (const a8::XParams& param)
                       {
                           Entity* entity = (Entity*)param.sender.GetUserData();
                           switch (entity->entity_type) {
                           case ET_Loot:
                               break;
                           case ET_Player:
                               {
                                   entity->room->moveable_hash_.erase(entity->entity_uniid);
                                   entity->room->human_hash_.erase(entity->entity_uniid);
                               }
                               break;
                           default:
                               {
                                   abort();
                               }
                               break;
                           }
                           entity->room->uniid_hash_.erase(entity->entity_uniid);
                           delete entity;
                       };
    xtimer.AddDeadLineTimerAndAttach(0,
                                     a8::XParams()
                                     .SetSender(entity),
                                     remove_func,
                                     &entity->xtimer_attacher.timer_list_);
}

void Room::OnHumanDie(Human* hum)
{
    --alive_count_;
    --App::Instance()->perf.alive_count;
    NotifyUiUpdate();
}

bool Room::OverBorder(const a8::Vec2 pos, float radius)
{
    //检查x轴
    {
        int left_x = pos.x - radius;
        if (left_x < 0.001f) {
            return true;
        }
        int right_x = pos.x + radius;
        if (right_x > map_meta->p->map_width()) {
            return true;
        }
    }
    //检查y轴
    {
        int up_y = pos.y + radius;
        if (up_y > map_meta->p->map_height()) {
            return true;
        }
        int down_y = pos.y - radius;
        if (down_y < 0.001f) {
            return true;
        }
    }
    return false;
}

Human* Room::GetWatchWarTarget(Human* hum)
{
    if (hum->team_members) {
        for (auto& member : *hum->team_members) {
            if (member != hum && !member->dead) {
                return member;
            }
        }
    }
    std::vector<Human*> humans;
    for (auto& pair : human_hash_) {
        if (pair.first != hum->entity_uniid && !pair.second->dead) {
            humans.push_back(pair.second);
        }
    }
    if (!humans.empty()) {
        Human* target = humans[rand() % humans.size()];
        return target;
    }
    return nullptr;
}

bool Room::BattleStarted()
{
    return battle_start_frameno_ != 0;
}

int Room::GetAliveTeamNum()
{
    int num = 0;
    for (auto& pair : team_hash_) {
        for (Human* hum : pair.second) {
            if (!hum->real_dead) {
                ++num;
                break;
            }
        }
    }
    return num;
}

bool Room::CanJoin(const std::string& accountid)
{
    if (gas_data.gas_mode != GasInactive) {
        return false;
    }
    if (accountid_hash_.find(accountid) != accountid_hash_.end()) {
        return false;
    }
    if (App::Instance()->HasFlag(5)) {
        return accountid_hash_.size() < 1;
    } else {
        return accountid_hash_.size() < ROOM_MAX_PLAYER_NUM;
    }
}

void Room::OnPlayerOffline(Player* hum)
{
    bool has_player = false;
    for (auto& pair : accountid_hash_) {
        if (pair.second->socket_handle != 0) {
            has_player = true;
        }
    }
    if (!has_player) {
        RoomMgr::Instance()->AddOverRoom(room_uuid);
    }
}

Entity* Room::FindFirstCollisonEntity(const a8::Vec2& aabb_pos, AabbCollider& aabb_box)
{
    Entity* target = nullptr;
    return target;
}

void Room::FindLocationWithAabb(Entity* target, const a8::Vec2& aabb_pos, AabbCollider* aabb_box,
                                float& new_x, float& new_y)
{
    a8::Vec2 old_pos = aabb_pos;
    a8::Vec2 new_pos = aabb_pos;
    AabbCollider target_collider;
    target->GetAabbBox(target_collider);
    {
        bool ret = aabb_box->CalcSafePointEx(aabb_pos, &target_collider, new_pos);
        if (!ret) {
            abort();
        }
    }
    a8::Vec2 new_pos_dir = new_pos - old_pos;
    new_pos_dir.Normalize();
    float distance = (new_pos - old_pos).Norm();
    for (int i = distance; i < 100000; i += 5) {
        a8::Vec2 pos = old_pos + new_pos_dir * i;
        new_x = pos.x;
        new_y = pos.y;

        bool is_collision = false;
        std::set<ColliderComponent*> colliders;
        map_service.GetColliders(pos.x, pos.y, colliders);
        for (ColliderComponent* collider : colliders) {
            if (collider->IntersectEx(pos, aabb_box)) {
                is_collision = true;
                break;
            }
        }
        if (!is_collision) {
            break;
        }
    }
}

void Room::FillSMUiUpdate(cs::SMUiUpdate& msg)
{
}

std::set<Human*>* Room::GetAliveTeam()
{
    for (auto& pair : team_hash_) {
        for (Human* hum : pair.second) {
            if (!hum->real_dead) {
                return &pair.second;
            }
        }
    }
    return nullptr;
}

int Room::NewTeam()
{
    ++current_teamid;
    team_hash_[current_teamid] = std::set<Human*>();
    return current_teamid;
}

void Room::TouchPlayerList(a8::XParams param,
                           std::function<void (Player*, a8::XParams&)> func)
{
    if (!func) {
        return;
    }
    for (auto& pair : accountid_hash_) {
        if (pair.second) {
            func(pair.second, param);
        }
    }
}

void Room::TouchHumanList(a8::XParams param,
                          std::function<bool (Human*, a8::XParams&)> func)
{
    if (!func) {
        return;
    }
    for (auto& pair : human_hash_) {
        if (pair.second) {
            if (!func(pair.second, param)) {
                break;
            }
        }
    }
}

void Room::TouchEntityList(a8::XParams param,
                           std::function<bool (Entity*, a8::XParams&)> func)
{
    if (!func) {
        return;
    }
    for (auto& pair : uniid_hash_) {
        if (pair.second) {
            if (!func(pair.second, param)) {
                break;
            }
        }
    }
}

void Room::UpdateGas()
{
    switch (gas_data.gas_mode) {
    case GasInactive:
        UpdateGasInactive();
        break;
    }
}

void Room::UpdateGasInactive()
{
    if (frame_no - gas_data.gas_start_frameno >=
        MetaMgr::Instance()->gas_inactive_time * SERVER_FRAME_RATE) {
        #if 0
        gas_data.gas_mode = GasJump;
        #endif
        gas_data.gas_start_frameno = frame_no;
        if (human_hash_.size() < ROOM_MAX_PLAYER_NUM) {
            CreateAndroid(ROOM_MAX_PLAYER_NUM - human_hash_.size());
            NotifyUiUpdate();
        }
        CombineTeam();
        NotifyWxVoip();
        RoomMgr::Instance()->ActiveRoom(room_uuid);
    }
}

void Room::MatchTeam(Human* hum)
{
    for (auto& pair : human_hash_) {
        if (pair.second != hum) {
            if (!hum->team_uuid.empty() && pair.second->team_uuid == hum->team_uuid) {
                if (!pair.second->team_members) {
                    pair.second->team_id = NewTeam();
                    pair.second->team_members = &team_hash_[pair.second->team_id];
                    pair.second->team_members->insert(pair.second);
                }
                if (pair.second->team_members->size() < MAX_TEAM_NUM) {
                    pair.second->team_members->insert(hum);
                    hum->team_id = pair.second->team_id;
                    hum->team_members = pair.second->team_members;
                }
                break;
            }
        }
    }
    if (hum->team_id == 0) {
        hum->team_id = NewTeam();
        hum->team_members = &team_hash_[hum->team_id];
        hum->team_members->insert(hum);
    }
}

void Room::CombineTeam()
{
    std::map<int, size_t> need_combine_teams;
    std::map<int, size_t> need_combine_teams_copy;

    for (auto& pair : team_hash_) {
        for (Human* hum : pair.second) {
            if (hum->auto_fill && !hum->team_uuid.empty()) {
                need_combine_teams[pair.first] = pair.second.size();
                break;
            }
        }
    }

    need_combine_teams_copy = need_combine_teams;
    bool combine_ok = false;
    do {
        combine_ok = false;
        for (auto& pair1 : need_combine_teams) {
            for (auto& pair2 : need_combine_teams_copy) {
                if (pair1.first != pair2.first) {
                    if (pair1.second + pair2.second <= MAX_TEAM_NUM) {
                        int team_id1 = pair1.first;
                        int team_id2 = pair2.first;

                        {
                            std::set<Human*>& team1_members = team_hash_[team_id1];
                            std::set<Human*>& team2_members = team_hash_[team_id2];
                            if (pair1.second > pair2.second) {
                                for (Human* hum : team2_members) {
                                    hum->team_id = team_id1;
                                    hum->team_members = &team1_members;
                                    team1_members.insert(hum);
                                }
                                team_hash_.erase(pair2.first);
                            } else {
                                for (Human* hum : team1_members) {
                                    hum->team_id = team_id2;
                                    hum->team_members = &team2_members;
                                    team2_members.insert(hum);
                                }
                                team_hash_.erase(pair1.first);
                            }
                        }

                        need_combine_teams.erase(team_id1);
                        need_combine_teams.erase(team_id2);
                        need_combine_teams_copy.erase(team_id1);
                        need_combine_teams_copy.erase(team_id2);
                        combine_ok = true;
                        break;
                    }
                }
            }//end for
            if (combine_ok) {
                break;
            }
        }
    } while (combine_ok);
}

void Room::AddObjectLater(Entity* entity)
{
    auto add_func = [] (const a8::XParams& param)
                       {
                           Entity* entity = (Entity*)param.sender.GetUserData();
                           entity->room->uniid_hash_[entity->entity_uniid] = entity;
                           entity->room->later_add_hash_.erase(entity->entity_uniid);
                       };
    later_add_hash_[entity->entity_uniid] = entity;
    xtimer.AddDeadLineTimerAndAttach(0,
                                     a8::XParams()
                                     .SetSender(entity),
                                     add_func,
                                     &entity->xtimer_attacher.timer_list_);
}

void Room::OnGameOver()
{
    for (auto& pair : human_hash_) {
        if (game_over && game_over_frameno == frame_no) {
            pair.second->SendGameOver();
        }
    }
    RoomMgr::Instance()->AddOverRoom(room_uuid);
}

void Room::RandRemoveAndroid()
{
    Human* hum = nullptr;
    for (auto& pair : human_hash_) {
        if (pair.second->entity_subtype == EST_Android) {
            hum = pair.second;
            break;
        }
    }
    if (hum) {
        if (hum->team_id != 0) {
            team_hash_.erase(hum->team_id);
        }
        for (auto& cell : hum->grid_list) {
            for (Human* target : cell->human_list) {
                target->AddOutObjects(hum);
            }
            cell->human_list.erase(hum);
        }
        moveable_hash_.erase(hum->entity_uniid);
        uniid_hash_.erase(hum->entity_uniid);
        human_hash_.erase(hum->entity_uniid);
        removed_robot_hash_[hum->entity_uniid] = hum;
        --alive_count_;
        --App::Instance()->perf.alive_count;
        for (auto& pair : human_hash_) {
            pair.second->RemovePartObjects(hum);
        }
    }
}

void Room::NotifyUiUpdate()
{
    xtimer.AddDeadLineTimerAndAttach(0,
                                     a8::XParams()
                                     .SetSender(this),
                                     [] (const a8::XParams& param)
                                     {
                                         Room* room = (Room*)param.sender.GetUserData();
                                         room->TouchPlayerList(a8::XParams(),
                                                               [] (Player * hum, a8::XParams & param)
                                                               {
                                                                   hum->SendUIUpdate();
                                                               });
                                     },
                                     &xtimer_attacher.timer_list_);
}

void Room::NotifyWxVoip()
{
    xtimer.AddDeadLineTimerAndAttach(0,
                                     a8::XParams()
                                     .SetSender(this),
                                     [] (const a8::XParams& param)
                                     {
                                         Room* room = (Room*)param.sender.GetUserData();
                                         room->TouchPlayerList(a8::XParams(),
                                                               [] (Player * hum, a8::XParams & param)
                                                               {
                                                                   hum->SendWxVoip();
                                                               });
                                     },
                                     &xtimer_attacher.timer_list_);
}

bool Room::GenSnakeBornPos(a8::Vec2& pos)
{
    return true;
}
